header {* Lemmas on Connection *}

theory Connected imports MessagesAndPaths
begin

subsection{*a few lemmas on intersects*}
(*improve later*)
lemma intersect_node_simp: "intersects Z (CAN_Zones C N) = intersects (CAN_Zones C N) Z"
apply (auto simp: intersects_def)
done

lemma intersects_union[simp]: "intersects Z (Z'\<union>Z'') = (intersects Z Z' \<or> intersects Z Z'')"
apply (auto simp: intersects_def)
done

lemma intersects_Union[simp]: "intersects Z (\<Union>ZSet) = (\<exists> Z'\<in>ZSet. intersects Z Z')"
apply (auto simp: intersects_def)
done

lemma intersects_Union_list: "\<lbrakk>L\<le> length list\<rbrakk>\<Longrightarrow>intersects Z (\<Union>i<L. f(list!i)) = (\<exists> i<L. intersects Z (f(list!i)))"
apply (auto simp: intersects_def)
done

lemma NodesInZone_Union[simp]: "NodesInZone C (\<Union>ZSet) = \<Union>{ NodesInZone C Z'|Z'. Z'\<in>ZSet}"
apply (auto simp: intersects_def)
done


subsection{*Lemmas on Connected *}

text{*MARK*}
lemma Connected_exists_neighbour: 
 "\<lbrakk>Connected C Z; ZoneSize C Z>1\<rbrakk>
  \<Longrightarrow>\<forall> N\<in>NodesInZone C Z. \<exists> N'\<in>NodesInZone C Z. CAN_neighbour C N N'"
apply (case_tac "ZoneSize C Z")
apply simp;
apply (case_tac nat);
apply auto;
apply (simp add: Finite_Set.card_Suc_eq);
apply (elim exE conjE);
apply (unfold Connected_def)
apply (drule_tac x=N in bspec)
apply force
apply (rotate_tac 2, drule sym)
apply (subgoal_tac "\<exists> nn. nn\<noteq>N\<and>nn\<in>NodesInZone C Z")
apply (elim exE)
apply (drule_tac x=nn in bspec)
apply simp+
apply clarify
apply (subgoal_tac "length node_list>1")
apply (rule_tac x="node_list!1" in exI)
apply (case_tac "length node_list = 2")
apply simp+
apply (case_tac "length node_list")
apply force
apply (case_tac nat)
apply auto
apply (case_tac "b=N")
apply blast
apply blast
apply (case_tac "b=N")
apply blast
apply blast
done

(*Connected if not distinct sufficient*)

lemma NodeList_Distinct[rule_format]:
"\<forall> n. (node_list\<noteq>[]\<longrightarrow> node_list!0=n \<longrightarrow> destination_NL node_list=n' 
         \<longrightarrow> (\<forall> i<length node_list - 1. node_list!i\<in>CAN_Nodes C
              \<and> CAN_neighbour C (node_list!i) (node_list!(i+1)) 
              \<and> node_list!i\<in>NodesInZone C Z) \<longrightarrow> 
(\<exists> node_list. distinct node_list \<and>node_list\<noteq>[]\<and> node_list!0=n \<and> destination_NL node_list=n' 
         \<and> (\<forall> i<length node_list - 1. node_list!i\<in>CAN_Nodes C
              \<and> CAN_neighbour C (node_list!i) (node_list!(i+1)) 
              \<and> node_list!i\<in>NodesInZone C Z)))"
apply (induct_tac node_list)
apply simp+
apply clarify
apply (case_tac list)
apply clarsimp
apply (rule_tac x="[a]" in exI)
apply force
apply clarsimp
apply (erule impE)
apply force
apply clarsimp
apply (case_tac "a\<in>set node_list")
apply (simp add: set_conv_nth)
apply clarsimp
apply (rule_tac x="drop i node_list" in exI)
apply simp+
apply (rule_tac x="a#node_list" in exI)
apply (case_tac "length node_list")
apply simp+
apply clarsimp
apply (case_tac i)
apply simp+
apply (drule_tac x=0 in spec)+
apply simp+
done

lemma Connected_path_inside_zone[rule_format]: "\<lbrakk>Connected C Z ; (ZoneSize C Z)>1\<rbrakk> \<Longrightarrow> 
\<exists> msgs. finite msgs \<and> msgs\<noteq>{} \<and> 
(\<forall> N\<in> NodesInZone C Z. \<forall> N'\<in> NodesInZone C Z.  N\<noteq>N'\<longrightarrow> 
   (\<exists> MsgP. MsgP\<in> path_inside_zone C msgs Z\<and> source MsgP = N \<and> destination MsgP = N'))"
 --" for a connected zone, there is a finite set of messages (msgs) such that any two nodes can be joined by a message path inside msgs"

apply (unfold Connected_def)
apply auto;
apply (rule_tac x="{<0|s,d,{tup. \<forall>d. Rep_Space (tup d) \<le> SIZE_Space}>|s d. s\<in> CAN_Nodes C \<and>d \<in> CAN_Nodes C \<and>CAN_neighbour C s d}" in exI)
apply auto;
(*3*)
  apply (subgoal_tac "{< 0| s, d, {tup. \<forall>d. Rep_Space (tup d) \<le> SIZE_Space}> |s d. s \<in> CAN_Nodes C \<and> d \<in> CAN_Nodes C\<and>CAN_neighbour C s d}\<subseteq>{< 0| s, d, {tup. \<forall>d. Rep_Space (tup d) \<le> SIZE_Space}> |s d. s \<in> CAN_Nodes C  \<and> d \<in> CAN_Nodes C}")
   apply (rule finite_subset)
    apply simp+
   apply (subgoal_tac "finite ((CAN_Nodes C)\<times>(CAN_Nodes C))")
     apply (subgoal_tac "{< 0| s, d, {tup. \<forall>d. Rep_Space (tup d) \<le> SIZE_Space}> |s d. s \<in> CAN_Nodes C \<and> d \<in> CAN_Nodes C} = 
                        (\<lambda> sd. < 0| fst sd, snd sd, {tup. \<forall>d. Rep_Space (tup d) \<le> SIZE_Space}>) ` (CAN_Nodes C \<times> CAN_Nodes C)")
      apply (drule_tac h="\<lambda> sd. < 0| fst sd, snd sd,{tup. \<forall>d. Rep_Space (tup d) \<le> SIZE_Space} >" in  Finite_Set.finite_imageI)
      apply simp
      apply auto
(*4*)
  apply (subgoal_tac "< 0| s, d, {tup. \<forall>d. Rep_Space (tup d) \<le> SIZE_Space}> = (\<lambda>sd. < 0| fst sd, snd sd,{tup. \<forall>d. Rep_Space (tup d) \<le> SIZE_Space} >) (s,d)")
   apply (rotate_tac -1,erule ssubst);
   apply (rule imageI)
   apply auto
 apply (insert finite_CAN_Nodes[of C]);
 apply force;
defer

(*2*)
 apply (drule_tac x=N in spec)
  apply simp
 apply (drule_tac x=N' in spec)
  apply simp+;
 apply clarify;
 apply (subgoal_tac "\<exists> MsgP. length MsgP =length node_list - Suc 0\<and> (\<forall> i<length MsgP. MsgP!i =<0| node_list!i,node_list!Suc(i),{tup. \<forall>d. Rep_Space (tup d) \<le> SIZE_Space}>)")
  apply auto
  apply (rule_tac x=MsgP in exI)
  apply auto
    apply (simp add: path_inside_zone_def valid_path_def)
    apply auto
    
(*11*)
    apply (case_tac "i<length node_list - 2")
     apply force
    apply (subgoal_tac "i =  length node_list - 2")
     apply simp
     apply (subgoal_tac " Suc (length node_list - 2) = (length node_list - 1)")
      apply simp+
(*10*)
   apply (simp add: Message_source_def Message_dest_def)
  apply (force simp: List.distinct_conv_nth)
 apply (simp add: Message_source_def)
 apply (rotate_tac -1,drule_tac x="0" in spec)
 apply force

(*7*)
 apply (rotate_tac -1,drule_tac x="0" in spec)
 apply simp+
 apply (case_tac "length node_list")
  apply force
 apply simp
 apply (case_tac nat)
  apply simp+
 apply (simp add: Message_source_def Message_dest_def)+
apply (case_tac "i<length node_list - 2")
 apply (force simp: Message_dest_def intersects_def)
apply (subgoal_tac "i =  length node_list - 2")
(*7*)
  apply simp
  apply (subgoal_tac " Suc (length node_list - 2) = (length node_list - 1)")
   apply simp+
apply (simp add: Message_source_def Message_dest_def)
apply (case_tac "i<length node_list - 2")
 apply (force simp: Message_dest_def intersects_def)
apply (subgoal_tac "i =  length node_list - 2")
 apply simp
 apply (subgoal_tac " Suc (length node_list - 2) = (length node_list - 1)")
  apply simp+
(*4*)
  apply (force simp: Message_source_def)
 apply (rotate_tac -1,drule_tac x="(length node_list - Suc (Suc 0))" in spec)
 apply (case_tac "length node_list")
  apply force
(*3*)
 apply simp+
 apply (case_tac nat)
  apply simp+
 apply (simp add: Message_dest_def)
apply (subgoal_tac "(Suc 0 < (ZoneSize C Z)\<and> node_list ! 0 \<in> CAN_Nodes C\<and> node_list ! (length node_list - Suc 0) \<in> CAN_Nodes C\<and>     node_list ! 0 \<noteq> node_list ! (length node_list - Suc 0)\<and> intersects (CAN_Zones C (node_list ! 0)) Z\<and>
        intersects (CAN_Zones C (node_list ! (length node_list - Suc 0))) Z\<and> finite (CAN_Nodes C)\<and> distinct node_list\<and>
        (\<forall>i<length node_list - Suc 0.
           node_list ! i \<in> CAN_Nodes C \<and> CAN_neighbour C (node_list ! i) (node_list ! Suc i) \<and> intersects (CAN_Zones C (node_list ! i)) Z)) 
\<longrightarrow>
       ( \<exists>MsgP. length MsgP = length node_list - Suc 0 \<and>
                 (\<forall>i<length MsgP. MsgP ! i = < 0| node_list ! i, node_list ! Suc i, {tup. \<forall>d. Rep_Space (tup d) \<le> SIZE_Space}>))")
 apply force
apply (thin_tac "Suc 0 < card {N\<in> CAN_Nodes C. intersects (CAN_Zones C N) Z}")
apply (thin_tac" intersects (CAN_Zones C (node_list ! (length node_list - Suc 0))) Z")
apply (thin_tac "node_list ! 0 \<noteq> node_list ! (length node_list - Suc 0)")
apply (thin_tac " node_list ! (length node_list - Suc 0) \<in> CAN_Nodes C")

apply (induct_tac node_list)
 apply force;
apply clarify;
apply (case_tac list);
 apply force
apply clarify;
apply (case_tac lista);
 apply simp+;
 apply (rule_tac x="[<0|a,aa,{tup. \<forall>d. Rep_Space (tup d) \<le> SIZE_Space}>]" in exI)
 apply simp+
(*2*)
 apply (erule impE)
  apply rule
   apply (drule_tac x="Suc 0" in spec);
   apply force
  apply (rule, simp add: distinct_conv_nth);
   apply clarify;
   apply (case_tac "length listb")
    apply simp+;
  apply rule
   apply force
  apply clarify;
  apply rule;
   apply force;
  apply force;
 apply clarify;
 apply (rule_tac x="<0|a,aa,{tup. \<forall>d. Rep_Space (tup d) \<le> SIZE_Space}>#MsgP" in exI)
 apply simp+;
 apply clarify;
 apply (case_tac i);
  apply force;
 apply force;
apply (subgoal_tac "Connected C Z");
 apply (drule Connected_exists_neighbour)
  apply (auto simp: Connected_def)
apply (insert non_singleton_rev)
apply (drule_tac x="NodesInZone C Z" in meta_spec)
apply auto
apply (rotate_tac 1,drule_tac x=a in spec)
apply auto
done

(*inside a connected zone, there are paths inside the zone*)
lemma path_set_Connected_non_empty: 
"\<lbrakk>Connected C Z;  (ZoneSize C Z)>1\<rbrakk> \<Longrightarrow> \<exists> msgs. path_inside_zone C msgs Z\<noteq>{}"

apply (drule Connected_path_inside_zone);
 apply simp+;
apply clarify;
apply (case_tac  "ZoneSize C Z")
 apply simp+;
apply (simp add: Finite_Set.card_Suc_eq);
apply clarify;
apply (case_tac  "card B")
 apply simp+;
apply (simp add: Finite_Set.card_Suc_eq);
apply clarify;
apply (drule_tac x=b in spec)
apply (rotate_tac -1, erule impE)
 apply (simp add: intersects_def)
 apply auto
done


lemma connected_union_pre:"Suc 0<ZoneSize C (Z\<union>Z')\<Longrightarrow>  (Z={}\<or>Z'={} \<or> (\<exists>N\<in>NodesInZone C Z. \<exists> N' \<in> NodesInZone C Z'. N\<noteq>N'))"
apply (subgoal_tac "ZoneSize C (Z\<union>Z')>1")
apply (drule non_singleton_rev)
apply (auto simp: intersect_node_simp)
apply (drule_tac C=C in intersects_empty_rev, auto simp: intersect_node_simp)+
done

lemma connected_union:"\<lbrakk>Connected C (Z \<union> Z'); Suc 0<ZoneSize C (Z\<union>Z')\<rbrakk>\<Longrightarrow>  (Z={}\<or>Z'={} \<or> (\<exists>N\<in>NodesInZone C Z.\<exists>N'\<in>NodesInZone C Z'. CAN_neighbour C N N'))"
apply (drule connected_union_pre)
apply auto
apply (simp add: Connected_def)
apply (drule_tac x=N in spec)
apply simp+
apply (rotate_tac -1,drule_tac x=N' in spec)
apply auto
apply (subgoal_tac "\<exists> i<length node_list - Suc 0. intersects (CAN_Zones C (node_list ! i)) Z \<and> intersects (CAN_Zones C (node_list ! (Suc i))) Z' ")
apply clarify
apply (drule_tac x="(node_list ! i)" in spec)
apply simp+
apply (rotate_tac -1,drule_tac x="(node_list ! Suc i)" in spec)
apply (case_tac "i=length node_list-Suc (Suc 0)")
apply auto
apply (subgoal_tac "Suc ((length node_list - Suc (Suc 0))) = length node_list - (Suc 0)")
apply auto
apply (case_tac "length node_list")
apply simp+
apply (case_tac nat)
apply simp+

apply (subgoal_tac  "(\<forall>i<Suc nata.
           node_list ! i \<in> CAN_Nodes C \<and>
           CAN_neighbour C (node_list ! i) (node_list ! Suc i) \<and>
           node_list ! i \<in> CAN_Nodes C \<and> (intersects (CAN_Zones C (node_list ! i)) Z \<or> intersects (CAN_Zones C (node_list ! i)) Z'))\<and>intersects (CAN_Zones C (node_list ! 0)) Z\<and> intersects (CAN_Zones C (node_list ! Suc nata)) Z' \<longrightarrow>(\<exists>i<Suc nata. intersects (CAN_Zones C (node_list ! i)) Z \<and> intersects (CAN_Zones C (node_list ! Suc i)) Z')")
apply force
apply (induct_tac nata)
apply simp+
apply clarify
apply (case_tac "intersects (CAN_Zones C (node_list ! Suc n)) Z'")
apply force
apply (rule_tac x="Suc n" in exI)
apply auto
done

lemma connected_union_rev:"\<lbrakk>Connected C Z; Connected C Z'; N\<in>NodesInZone C Z;N'\<in>NodesInZone C Z'; CAN_neighbour C N N'\<rbrakk>\<Longrightarrow>Connected C (Z \<union> Z')  "
apply (simp add: Connected_def)
apply auto
apply (drule_tac x=n in spec,simp)
apply (rotate_tac -1,drule_tac x=n' in spec,simp)
apply clarify
apply (rule_tac x=node_list in exI)
apply auto

prefer 3

apply (rotate_tac 1)
apply (drule_tac x=n in spec,simp)
apply (rotate_tac -1,drule_tac x=n' in spec,simp)
apply clarify
apply (rule_tac x=node_list in exI)
apply auto

(*2*)
apply (drule_tac x=n in spec,simp)
apply (rotate_tac -1,drule_tac x=N in spec,simp)
apply (drule_tac x=N' in spec,simp)
apply (rotate_tac -1,drule_tac x=n' in spec,simp)
apply clarify
apply (subgoal_tac "node_list@node_lista\<noteq>[]")
apply (drule_tac node_list="node_list@node_lista" and C=C and Z="Z\<union>Z'" in NodeList_Distinct )
apply simp+
apply (simp add: nth_append)
apply (case_tac "i=(length node_list - Suc 0)")
apply clarsimp+
apply rule
apply force
apply clarsimp
apply (rotate_tac -4, drule_tac x="(i - length node_list)" in spec)
apply (erule impE)
apply arith
apply clarsimp
apply (subgoal_tac "Suc (i - length node_list) = (Suc i - length node_list)")
apply simp
apply arith

(*2*)
apply clarsimp
apply (rule_tac x="node_listb" in exI)
apply simp+
apply (simp add: nth_append)
apply (case_tac "node_list")
apply simp+

(*1 - symmetric*)

apply (drule neighbour_refl)

apply (drule_tac x=N in spec,simp)
apply (rotate_tac -1,drule_tac x=n' in spec,simp)
apply (drule_tac x=n in spec,simp)
apply (rotate_tac -1,drule_tac x=N' in spec,simp)
apply clarify
apply (subgoal_tac "node_lista@node_list\<noteq>[]")
apply (drule_tac node_list="node_lista@node_list" and C=C and Z="Z\<union>Z'" in NodeList_Distinct )
apply simp+
apply (simp add: nth_append)
apply (case_tac "i=(length node_lista - Suc 0)")
apply clarsimp+
apply rule
apply force
apply clarsimp
apply (rotate_tac -6, drule_tac x="(i - length node_lista)" in spec)
apply (erule impE)
apply arith
apply clarsimp
apply (subgoal_tac "Suc (i - length node_lista) = (Suc i - length node_lista)")
apply simp
apply arith

(*2*)
apply clarsimp
apply (rule_tac x="node_listb" in exI)
apply simp+
apply (simp add: nth_append)
apply (case_tac "node_list")
apply simp+
done

lemma node_list_neighbour[rule_format]:"
 node_list!0\<in>NodesInZone C Z \<longrightarrow> destination_NL node_list\<in> NodesInZone C Z'
  \<longrightarrow> length node_list>1 \<longrightarrow>
  (\<forall> i<length node_list - 1. node_list!i\<in>CAN_Nodes C
     \<and> CAN_neighbour C (node_list!i) (node_list!(i+1)) 
     \<and> node_list!i\<in>NodesInZone C (Z\<union>Z'))\<longrightarrow>
(\<exists> N\<in>CAN_Nodes C. \<exists> N'\<in>CAN_Nodes C.  (N\<in>NodesInZone C Z\<and>N'\<in>NodesInZone C Z'\<and>CAN_neighbour C N N'))"
apply (induct_tac node_list)
apply force
apply clarify
apply simp+
apply (case_tac list)
apply simp+
apply (case_tac lista)
apply force
apply simp+
apply (case_tac "aa\<in>NodesInZone C Z")
apply (erule impE)
apply force
apply (erule impE)
apply simp+
apply clarify
apply simp+

apply (drule_tac x="Suc i" in spec)
apply force

apply assumption

apply (subgoal_tac "aa\<in>NodesInZone C Z'")
apply (rule_tac x=a in bexI)
apply simp+
apply (rule_tac x=aa in bexI)
apply simp+
apply force
apply force
apply force
apply (drule_tac x=1 in spec)+
apply force
done

lemma map_Union[intro]: " N\<le>length list \<Longrightarrow>(\<Union> i<N . F (list ! i)) = (\<Union> i<N. (map F list) ! i)"
by auto

lemma Union_unfold[rule_format]: "\<forall> N. N\<le>length list \<longrightarrow>(\<Union> i<Suc N. (a#list) ! i) = (\<Union> i<N. list ! i)\<union> a"
apply(induct_tac list)
apply auto
apply (case_tac xa)
apply auto
done



lemma not_connected_union_not_empty: "\<lbrakk>\<not> Connected C (Z\<union>Z');Connected C Z; Connected C Z'\<rbrakk>\<Longrightarrow>Z'\<noteq>{}"
apply (simp add: Connected_def intersects_def)
apply auto
apply (drule_tac x=n in spec)
apply simp+
apply (erule impE)
apply force
apply (drule_tac x=n' in spec)
apply blast
done
lemma Connected_union_smallZones: "\<lbrakk>NodesInZone C (Z\<union>Z')={N};Z\<noteq>{}\<rbrakk>\<Longrightarrow>Connected C (Z\<union>Z'\<union>Z'')=Connected C (Z\<union>Z'')"
apply (case_tac "Z'={}")
apply simp
apply (subgoal_tac "\<forall> N1\<in>CAN_Nodes C. (intersects (CAN_Zones C N1) Z \<longrightarrow>N1=N)")
apply (subgoal_tac "\<forall> N1\<in>CAN_Nodes C. intersects (CAN_Zones C N1) Z' = intersects (CAN_Zones C N1) Z")
defer
apply (subgoal_tac "\<forall> N1\<in>CAN_Nodes C. intersects (CAN_Zones C N1) Z' \<longrightarrow>N1=N")
apply clarify
apply (simp add: intersects_def)
apply rule
apply clarify
apply (subgoal_tac "\<exists>x. x\<in>Z")
apply clarify
apply (rotate_tac -1,frule_tac C=C in intersects_empty_rev)
apply clarify
apply (rule_tac x=x in exI)
apply simp
apply (subgoal_tac "node=N")
apply (subgoal_tac "N1=N")
apply blast
apply blast
apply blast
apply blast
apply (subgoal_tac "\<exists>x. x\<in>Z'")
apply clarify
apply (rotate_tac -3,frule_tac C=C in intersects_empty_rev)
apply clarify
apply (rule_tac x=x in exI)
apply simp
apply (subgoal_tac "node=N")
apply (subgoal_tac "N1=N")
apply blast
apply blast
apply blast
apply blast
apply (force simp: intersects_def)
apply (force simp: intersects_def)
apply auto
apply (simp add: Connected_def)
apply clarify
apply (drule_tac x=n in spec)
apply simp
apply (drule_tac x=n' in spec)
apply blast
apply (simp add: Connected_def)
apply clarify
apply (drule_tac x=n in spec)
apply simp
apply (drule_tac x=n' in spec)
apply blast
done

declare intersect_node_simp[simp]
lemma Connected_Union[rule_format]: "(\<forall>i<length ZL. Connected C (Z\<union>(ZL!i))) \<longrightarrow> Z\<noteq>{}\<longrightarrow>(\<forall>i<length ZL. Connected C (ZL!i)) \<longrightarrow>Connected C Z \<longrightarrow> Connected C (Z\<union>(\<Union>i<length ZL. (ZL!i)))"
apply (induct_tac ZL)
apply auto
apply (drule_tac x=0 in spec)
apply simp
apply (subgoal_tac "length list\<le>length list")
defer
apply force

apply (drule_tac a=a in Union_unfold,simp)
apply (case_tac "\<exists> N. NodesInZone C (Z\<union>a)={N}")
apply clarify
apply (drule_tac Z''="(\<Union> i<length list. list ! i)" in  Connected_union_smallZones)
apply blast
apply (subgoal_tac "(Z  \<union> ((\<Union> i<length list. list ! i)\<union> a)) = (Z \<union> a \<union> (\<Union> i<length list. list ! i))")
apply simp
apply blast

apply (subgoal_tac "Connected C ((Z \<union> (\<Union> x<length list. list ! x)) \<union> a)")
apply (simp add: Un_assoc)
apply (rotate_tac -3,frule connected_union)
defer
apply simp
apply safe
apply simp+
apply (rule_tac N=N and N'=xa in  connected_union_rev)
apply force
apply force
apply simp+
apply (subgoal_tac " finite {N \<in> CAN_Nodes C. intersects (CAN_Zones C N) Z \<or> intersects (CAN_Zones C N) a}")
apply (rotate_tac -3,frule_tac C=C in intersects_empty_rev)
apply clarify
apply (drule_tac a=node in Lemmas.finite_non_singleton)
apply simp+
apply (insert CAN.finite_CAN_Nodes[of C])
apply force
done

lemma SingleNodeConnected: "NodesInZone C Z = {N} \<Longrightarrow> Connected C Z"
apply (simp add: Connected_def)
apply (rule_tac x="[N]" in exI)
apply force
done



lemma Connected_union_Union[rule_format]:"\<lbrakk>YSet\<noteq>{};Connected C ((\<Union>YSet)\<union>Z) ; (\<forall>Y\<in>YSet. Connected C Y); Connected C Z\<rbrakk>\<Longrightarrow> 
(\<exists>Y\<in>YSet. Connected C (Y\<union>Z))"
apply (case_tac "Z={}")
apply force
apply (case_tac "\<Union>YSet={}")
apply force
apply clarsimp
apply (case_tac "Suc 0<ZoneSize C (\<Union>YSet\<union>Z)")
apply (drule_tac connected_union)
apply simp+
apply (rotate_tac -1, erule disjE)
apply blast
apply clarsimp
apply (subgoal_tac "\<exists>Y\<in>YSet. intersects (CAN_Zones C N) Y")
apply clarsimp
apply (rule_tac x=Y in bexI)

apply (rule_tac N=N and N'=xb in connected_union_rev)
apply simp+
apply (force simp: intersects_def)

apply (subgoal_tac "\<exists> N. NodesInZone C (\<Union>YSet\<union>Z)={N}")
apply clarify
apply (subgoal_tac "NodesInZone C (x\<union>Z)={N}")
apply (drule SingleNodeConnected)+
apply force
apply (subgoal_tac "NodesInZone C (x \<union> Z) \<subseteq> NodesInZone C (\<Union>YSet \<union> Z)")
apply rule
apply force

apply (simp add: intersects_def)
apply rule
apply force
apply (subgoal_tac "\<exists> tup. tup\<in>x")
apply clarsimp
apply (rule_tac x=tup in exI)
apply (frule_tac C=C in intersects_empty_rev)
apply clarsimp
apply (subgoal_tac "node=N")
apply simp
apply force
apply blast
apply (force simp: intersects_def)
apply (subgoal_tac "\<exists> tup. tup\<in>x")
apply clarsimp
apply (frule_tac C=C in intersects_empty_rev)
apply clarsimp
apply (subgoal_tac "node\<in>{N\<in> CAN_Nodes C. intersects (CAN_Zones C N) (\<Union>YSet)}\<or>intersects (CAN_Zones C N) Z")
 defer
apply (simp add: intersects_def)
apply rule
apply (rule_tac x=tup in exI)
apply force
apply blast
apply (rule_tac x=node in exI)
apply simp+
apply (insert CAN.finite_CAN_Nodes[of C])
apply (subgoal_tac "node\<in>{N \<in> CAN_Nodes C. intersects (CAN_Zones C N) (\<Union>YSet) \<or> intersects (CAN_Zones C N) Z}")
apply (case_tac "ZoneSize C (\<Union>YSet \<union> Z)")
apply force
apply simp
apply (simp add: Finite_Set.card_Suc_eq)
apply force
apply (force simp: intersects_def)
done

lemma Connected_union_Union_2[rule_format]:"\<lbrakk>Connected C (X\<union>(\<Union>YSet)\<union>Z) ; Connected C X; (\<forall>Y\<in>YSet. Connected C Y); Connected C Z\<rbrakk>\<Longrightarrow> 
(Connected C (X\<union>Z)\<or>(\<exists>Y\<in>YSet. Connected C (Y\<union>Z)))"
apply (insert Connected_union_Union[of "{X}\<union>YSet" C Z])
apply auto
done

lemma Connected_union_union:"\<lbrakk>Connected C ((X\<union>Y)\<union>Z) ; Connected C X; Connected C Y; Connected C Z\<rbrakk>\<Longrightarrow>(Connected C (X\<union>Z))\<or>Connected C (Y\<union>Z)"
apply (insert Connected_union_Union[of "{X,Y}" C Z])
apply auto
done


lemma not_connected_union_no_intersection: "\<lbrakk>\<not> Connected C (Z\<union>Z'); Connected C Z; Connected C Z'\<rbrakk>\<Longrightarrow> \<not>(\<exists> N. N\<in>NodesInZone C Z\<and>N\<in>NodesInZone C Z')"
apply (subgoal_tac "Z'\<noteq>{}")
apply (subgoal_tac "Z\<noteq>{}")
defer
apply (rule_tac Z=Z' and C= C in not_connected_union_not_empty)
apply (subst Un_commute)
apply simp+
apply (erule not_connected_union_not_empty)
apply simp
apply simp

apply (case_tac "ZoneSize C Z'")
apply (subgoal_tac "finite (NodesInZone C Z')")
apply (force simp: intersects_def)
apply (rule CAN.finite_intersects )

apply (case_tac nat)
apply (simp add: Finite_Set.card_Suc_eq)
apply clarify
apply simp+
apply (subgoal_tac "N=b")
apply simp+
apply (subgoal_tac "\<forall>N\<in>CAN_Nodes C. intersects (CAN_Zones C N) Z'\<longrightarrow>intersects (CAN_Zones C N) Z")
apply (simp add: Connected_def)
apply clarify
apply simp+
apply (subgoal_tac "\<forall>i<length node_list. node_list!i \<in>NodesInZone C Z")
apply (drule_tac x=n in spec)
apply (erule impE)
apply force
apply (rotate_tac -1, drule_tac x=n' in spec)
apply simp+
apply (erule impE)
apply force
apply clarify
apply (drule_tac x=node_lista in spec)
apply simp+
apply force
apply clarify
apply (case_tac "i=(length node_list) - 1")
apply simp
apply (drule_tac x=i in spec)+ 
apply force

apply clarify
apply (simp add: intersects_def)
apply (subgoal_tac "b=Na")
apply force

apply (force simp: intersects_def)
apply (force simp: intersects_def)

apply (frule_tac Z=Z' in Connected_exists_neighbour)
apply simp+

apply (erule contrapos_np)
apply simp+
apply clarify
apply (drule_tac x=N in spec)
apply clarsimp
apply (rule_tac N=N and N'=x in connected_union_rev)
apply simp+
done

lemma not_connected_union_no_path: "\<lbrakk>\<not> Connected C (Z\<union>Z'); Connected C Z; Connected C Z';N\<in>NodesInZone C Z; N'\<in>NodesInZone C Z'\<rbrakk>\<Longrightarrow> \<not>(\<exists> node_list. node_list\<noteq>[]\<and>node_list!0=N \<and> destination_NL node_list=N' \<and> distinct node_list
         \<and> (\<forall> i<length node_list - 1. node_list!i\<in>CAN_Nodes C
              \<and> CAN_neighbour C (node_list!i) (node_list!(i+1)) 
              \<and> node_list!i\<in>NodesInZone C (Z\<union>Z')))"
apply (case_tac "N=N'")

apply (drule_tac "not_connected_union_no_intersection")
apply simp
apply simp
apply simp

apply (erule contrapos_nn)
apply clarify
apply (subgoal_tac "(node_list ! 0) \<in>NodesInZone C Z")
apply (drule_tac Z'=Z'  in node_list_neighbour)
apply simp+

apply (case_tac node_list, simp+)
apply (case_tac list, simp+)

apply clarify
apply (rule_tac N=N and N'=N' in connected_union_rev)
apply simp+
done

declare intersect_node_simp[simp del]
lemma connected_Union_neighbour:
"\<lbrakk>Connected C (CAN_Zones C N \<union> (\<Union> i<length ZL. ZL ! i)); N \<in> CAN_Nodes C;
            \<forall>i<length ZL. Connected C (ZL ! i);
            \<forall>i<length ZL. \<forall>j<length ZL. i \<noteq> j \<longrightarrow> \<not> Connected C (ZL ! i \<union> ZL ! j); 
            i < length ZL;
            N'\<in>CAN_Nodes C; N'\<noteq>N;N'\<in>NodesInZone C (ZL ! i)\<rbrakk>
           \<Longrightarrow> \<exists>x. x \<in> CAN_Nodes C \<and> x\<in>NodesInZone C (ZL ! i) \<and> CAN_neighbour C N x"
apply (subgoal_tac "ZL!i\<noteq>{}")

apply (subgoal_tac "Connected C (CAN_Zones C N \<union> (\<Union> {ZL!j|j. j<length ZL\<and>j\<noteq>i}) \<union> ZL!i)")
apply simp+
apply (drule Connected_union_Union_2)
apply (rule_tac N=N in SingleNodeConnected)

apply (simp add: intersects_def)
apply rule
apply clarify
apply (rule_tac C=C in  CAN.inc_single_zone_2)
apply (simp del: intersect_node_simp)+
apply (force simp: intersects_def)
apply clarify
apply (drule CAN.zone_non_empty)+
apply force

apply force
apply force
apply safe
apply (rotate_tac -1,drule connected_union)
apply (simp)+
defer
apply simp+
apply (erule disjE)
apply (drule CAN.zone_non_empty)+
apply force
apply (erule disjE)

apply blast

apply clarify
apply (drule_tac N=Na and N'=N in CAN.inc_single_zone_2)
apply simp+
apply (subgoal_tac "\<Union>{(ZL ! j)|j. (j < length ZL \<and> j \<noteq> i)} \<union> ZL ! i = (\<Union> i<length ZL. ZL ! i)")
apply (force simp: Un_assoc)
apply blast
apply (case_tac ZL)
apply force
apply (case_tac list)
apply (force simp: intersects_def)
apply (case_tac i)
apply simp+
apply (rotate_tac 3,drule_tac x=1 in spec)
apply simp+
apply (rotate_tac -1,drule_tac x=0 in spec)
apply force
apply simp+
apply (rotate_tac 3,drule_tac x=0 in spec)
apply simp+
apply (rotate_tac -1,drule_tac x="Suc nat" in spec)
apply force

apply (subgoal_tac "N'\<in>NodesInZone C ((CAN_Zones C N)\<union>(ZL!i))")
apply (subgoal_tac "N\<in>NodesInZone C ((CAN_Zones C N)\<union>(ZL!i))")
apply clarify
apply (subgoal_tac "finite (NodesInZone C ((CAN_Zones C N)\<union>(ZL!i)))")
apply (case_tac "(ZoneSize C ((CAN_Zones C N)\<union>(ZL!i)))")
apply force
apply (case_tac nat)
apply (simp add: Finite_Set.card_Suc_eq)
apply clarify
apply (subgoal_tac "N\<in>{b}")
apply (subgoal_tac "N'\<in>{b}")
apply blast
apply blast
apply blast
apply simp+
defer

apply simp+
apply rule 
apply (drule CAN.zone_non_empty)+
apply (force simp: intersects_def)

apply (force simp: intersects_def)

apply (insert finite_CAN_Nodes)
apply force
done

end